'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var readableStream = require('readable-stream');
var events = require('events');

function noop () {}

function isFunction (value) {
  return typeof value === 'function'
}

function last (arr) {
  return arr.length > 0 ? arr[arr.length - 1] : undefined
}

function isPromise (value) {
  return value instanceof Promise
}

function once (fn) {
  let called = false;
  return function (err, data) {
    if (called) return
    called = true;
    fn(err, data);
  }
}

function supportConcurrency (
  concurrency,
  _transform,
  _flush,
  _destroy,
  isTransform = false
) {
  if (!Number.isInteger(concurrency) || concurrency <= 0) {
    throw new TypeError('Concurrency must be a positive integer')
  }
  if (typeof _transform !== 'function') {
    throw new TypeError(
      isTransform
        ? 'Expected transform method to be a function'
        : 'Expected write method to be a function'
    )
  }

  // Active jobs counter
  let jobs = 0;

  // First encountered error
  let error;

  // Require more chunks callback
  let next;

  // End stream callback
  let close;

  function transform (chunk, encoding, callback) {
    if (!error) {
      job.call(this, chunk, encoding);
    }

    if (!error && jobs < concurrency) {
      // Keep requiring more chunks
      callback();
    } else {
      // Save the current callback for jobs completion
      next = callback;
    }
  }

  function job (chunk, encoding) {
    jobs++;

    _transform.call(this, chunk, encoding, once((err, data) => {
      jobs--;

      error = error || err;

      if (!error && data !== undefined && isTransform) {
        this.push(data);
      }

      if (jobs <= 0) {
        const callback = close || next;
        if (callback) {
          callback(error);
        } else if (error) {
          this.emit('error', error);
        }
      } else if (!error && next) {
        const callback = next;
        next = undefined;
        callback();
      }
    }));
  }

  function flush (callback) {
    // Build close callback
    close = err => {
      if (err || !_flush) {
        callback(err);
      } else {
        _flush.call(this, callback);
      }
    };

    if (jobs <= 0) {
      close();
    }
  }

  function destroy (err, callback) {
    // Ensure error to stop data flowing
    error = error || err || new Error('Premature close');

    // Build close callback
    close = _destroy ? _destroy.bind(this, error, callback) : callback;

    if (jobs <= 0) {
      close();
    }
  }

  return [transform, flush, destroy]
}

function handlePromise (promise, callback) {
  promise
    .then(result => callback(null, result))
    .catch(callback);
}

function supportPromise (read) {
  return function (size, callback) {
    handlePromise(read.call(this, size), callback);
  }
}

function wrapReadMethod (read) {
  let sync = false;
  let reading = false;

  return function (size) {
    if (sync || reading) {
      if (sync) {
        read.call(this, size);
      }
      return
    }
    reading = true;

    const callback = once((err, data) => {
      if (err) {
        process.nextTick(() => this.emit('error', err));
        return
      }

      if (data !== undefined) {
        this.push(data);
      }

      if (this._readableState.ended) {
        return
      }

      reading = false;
      if (this._readableState.length < this._readableState.highWaterMark) {
        setImmediate(() => this._read(size));
      }
    });

    if (read.length >= 2) {
      if (isPromise(read.call(this, size, callback))) {
        callback(new Error('Overspecialized function'));
      }
      return
    }

    const out = read.call(this, size);
    if (isPromise(out)) {
      read = supportPromise(read);
      handlePromise(out, callback);
    } else {
      sync = true;
      callback();
    }
  }
}

function ensureCallback (fn, argsCount) {
  return function (...args) {
    const out = fn.apply(this, args.slice(0, argsCount + 1));
    const callback = args[argsCount];

    if (isPromise(out)) {
      callback(new Error('Overspecialized function'));
    }
  }
}

function ensurePromise (fn, argsCount) {
  return function (...args) {
    const out = fn.apply(this, args.slice(0, argsCount));
    const callback = args[argsCount];

    if (!isPromise(out)) {
      callback(new Error('Expected promise'));
      return
    }

    out
      .then(result => callback(null, result))
      .catch(callback);
  }
}

function wrapMethod (fn, argsCount = 0) {
  return fn.length > argsCount
    ? ensureCallback(fn, argsCount)
    : ensurePromise(fn, argsCount)
}

function ensureFunction (fn) {
  if (!isFunction(fn)) {
    throw new TypeError('Expected a function')
  }
  return fn
}

function supportPromises (options) {
  const { read, write, writev, final, destroy, flush, transform } = options;

  return {
    ...options,
    read: read ? wrapReadMethod(ensureFunction(read)) : read,
    write: write ? wrapMethod(ensureFunction(write), 2) : write,
    writev: writev ? wrapMethod(ensureFunction(writev), 1) : writev,
    final: final ? wrapMethod(ensureFunction(final)) : final,
    transform: transform ? wrapMethod(ensureFunction(transform), 2) : transform,
    flush: flush ? wrapMethod(ensureFunction(flush)) : flush,
    destroy: destroy ? wrapMethod(ensureFunction(destroy), 1) : destroy
  }
}

function voidRead () {
  this.push(null);
}

function voidWrite (chunk, encoding, callback) {
  callback(null);
}

function voidTransform (chunk, encoding, callback) {
  callback(null, chunk);
}

function handleConcurrency (options) {
  const { concurrency, transform, flush, destroy } = options;

  return concurrency
    ? supportConcurrency(concurrency, transform, flush, destroy, true)
    : [transform, flush, destroy]
}

function transform (options = {}) {
  options = supportPromises(options);

  const [transform, flush, destroy] = handleConcurrency(options);

  return new readableStream.Transform({
    ...options,
    transform: transform || voidTransform,
    flush,
    destroy
  })
}

function asString (chunks, encoding, callback) {
  let result = '';
  for (const chunk of chunks) {
    if (Buffer.isBuffer(chunk)) {
      result += chunk.toString(encoding);
    } else if (typeof chunk === 'string') {
      result += chunk;
    } else {
      return callback(new TypeError('Chunk must be buffer or string'))
    }
  }
  callback(null, result);
}

function asBuffer (chunks, callback) {
  let result = Buffer.from([]);
  for (const chunk of chunks) {
    if (Buffer.isBuffer(chunk)) {
      result = Buffer.concat([result, chunk]);
    } else if (typeof chunk === 'string') {
      result = Buffer.concat([result, Buffer.from(chunk)]);
    } else {
      return callback(new TypeError('Chunk must be buffer or string'))
    }
  }
  callback(null, result);
}

function asIs (chunks, callback) {
  callback(null, chunks);
}

function collect (target) {
  let chunks = [];

  return transform({
    objectMode: true,
    transform (chunk, encoding, callback) {
      if (target === undefined) {
        if (Buffer.isBuffer(chunk)) {
          target = 'buffer';
        } else if (typeof chunk === 'string') {
          target = 'utf8';
        } else {
          target = false;
        }
      }
      chunks.push(
        typeof chunk === 'string' && target === 'buffer'
          ? Buffer.from(chunk, encoding)
          : chunk
      );
      callback();
    },
    flush (callback) {
      if (target === 'buffer') {
        asBuffer(chunks, callback);
      } else if (typeof target === 'string') {
        asString(chunks, target, callback);
      } else {
        asIs(chunks, callback);
      }
      chunks = [];
    },
    destroy (err, callback) {
      chunks = [];
      callback(err);
    }
  })
}

function handleConcurrency$1 (options) {
  const { concurrency, write, final, destroy } = options;

  return concurrency
    ? supportConcurrency(concurrency, write, final, destroy)
    : [write, final, destroy]
}

function duplex (options = {}) {
  options = supportPromises(options);

  const [write, final, destroy] = handleConcurrency$1(options);
  const { read, writev } = options;

  return new readableStream.Duplex({
    ...options,
    read: read || voidRead,
    write: !write && !writev ? voidWrite : write,
    writev,
    final,
    destroy
  })
}

function eos (stream, options, callback) {
  if (isFunction(options)) {
    return eos(stream, {}, options)
  }
  if (!isFunction(callback)) {
    throw new TypeError('Expected callback')
  }
  return readableStream.finished(stream, options, callback)
}

function destroyStream (stream, err) {
  err = err || new Error('Premature close');

  if (stream.setHeader && isFunction(stream.abort)) {
    stream.abort();
  } else if (isFunction(stream.destroy)) {
    stream.destroy(err);
  } else {
    process.nextTick(() => stream.emit('error', err));
  }

  return err
}

function duplexify (readable, writable, options) {
  function read () {
    if (readable.readableFlowing !== null) {
      readable.resume();
      return
    }

    const listener = data => {
      if (!this.push(data)) {
        readable.pause();
      }
    };

    eos(readable, { writable: false }, err => {
      readable.removeListener('data', listener);

      if (err) {
        this.emit('error', err);
      } else {
        this.push(null);
      }
    });

    readable.addListener('data', listener);
  }

  function write (chunk, encoding, callback) {
    writable.write(chunk, encoding, callback);
  }

  function final (callback) {
    writable.end(callback);
  }

  function destroy (err, callback) {
    if (readable) {
      err = destroyStream(readable, err);
    }
    if (writable) {
      err = destroyStream(writable, err);
    }
    callback(err);
  }

  return duplex({
    ...options,
    read: readable ? read : undefined,
    write: writable ? write : undefined,
    writev: undefined,
    final: writable ? final : undefined,
    destroy
  })
}

function readStream (entry) {
  return Array.isArray(entry) ? entry[0] : entry
}

function readOptions (entry) {
  return Array.isArray(entry) ? entry[1] : {}
}

function toDestroyer (stream) {
  let destroyed = false;
  return function destroyer (err) {
    if (!stream.__closed__ && !destroyed) {
      destroyed = true;
      destroyStream(stream, err);
    }
  }
}

function compose (a, b) {
  return function composed (arg) {
    a(arg);
    b(arg);
  }
}

function buildDestroyer (entries) {
  return entries.map(readStream).map(toDestroyer).reduce(compose)
}

function handleInternal (entries, destroyOnError, callback) {
  const destroy = destroyOnError ? buildDestroyer(entries) : noop;

  let count = entries.length;
  let error = null;

  const end = err => {
    error = error || err;
    if (error) {
      destroy(error);
    }
    if (--count < 1) {
      callback(error);
    }
  };

  return entries
    .map(entry => {
      const stream = readStream(entry);
      const options = readOptions(entry);

      return eos(stream, options, err => {
        stream.__closed__ = true;
        end(err);
      })
    })
    .reduce(compose)
}

function finished (...args) {
  if (!isFunction(last(args))) {
    throw new TypeError('Expected callback')
  }
  if (args.length <= 1) {
    throw new Error('Expected at least one stream')
  }
  return handleInternal(args, false, args.pop())
}

function handle (...args) {
  if (!isFunction(last(args))) {
    throw new TypeError('Expected callback')
  }
  if (args.length <= 1) {
    throw new Error('Expected at least one stream')
  }
  return handleInternal(args, true, args.pop())
}

function isReadable (value) {
  return value instanceof events.EventEmitter && isFunction(value.push)
}

function isWritable (value) {
  return value instanceof events.EventEmitter && isFunction(value.write)
}

function isStream (value) {
  return isReadable(value) || isWritable(value)
}

function isDuplex (value) {
  return isReadable(value) && isWritable(value)
}

function isReadableStrictly (value) {
  return isReadable(value) && !isWritable(value)
}

function isWritableStrictly (value) {
  return isWritable(value) && !isReadable(value)
}

function readable (options = {}) {
  options = supportPromises(options);

  return new readableStream.Readable({
    ...options,
    read: options.read || voidRead
  })
}

function setStreamOptions (streams) {
  return streams.map(stream => [stream, { readable: true, writable: false }])
}

function mergeReadables (sources, options) {
  if (!Array.isArray(sources)) {
    throw new TypeError('Expected array of streams')
  }

  let listener;

  return readable({
    ...options,
    read () {
      if (!listener) {
        listener = data => {
          if (!this.push(data)) {
            for (const source of sources) {
              source.pause();
            }
          }
        };

        handle(...setStreamOptions(sources), err => {
          for (const source of sources) {
            source.removeListener('data', listener);
          }

          if (err) {
            this.emit('error', err);
          } else {
            this.push(null);
          }
        });

        for (const source of sources) {
          source.addListener('data', listener);
        }
      } else {
        for (const source of sources) {
          source.resume();
        }
      }
    },
    destroy (err, callback) {
      for (const source of sources) {
        err = destroyStream(source, err);
      }
      callback(err);
    }
  })
}

function handleConcurrency$2 (options) {
  const { concurrency, write, final, destroy } = options;

  return concurrency
    ? supportConcurrency(concurrency, write, final, destroy)
    : [write, final, destroy]
}

function writable (options = {}) {
  options = supportPromises(options);

  const [write, final, destroy] = handleConcurrency$2(options);
  const { writev } = options;

  return new readableStream.Writable({
    ...options,
    write: !write && !writev ? voidWrite : write,
    writev,
    final,
    destroy
  })
}

function setStreamOptions$1 (streams) {
  return streams.map(stream => [stream, { readable: false, writable: true }])
}

function mergeWritables (targets, options) {
  if (!Array.isArray(targets)) {
    throw new TypeError('Expected array of streams')
  }

  // Current write/final callback
  let next;

  // Current error
  let error;

  // Targets count with backpressure on
  let waiting;

  // Update current error and fire next callback if necessary
  function call (err) {
    error = err || error;
    if (next) {
      const callback = next;
      next = undefined;
      callback(error);
    }
  }

  function write (chunk, encoding, callback) {
    next = callback;

    // Handle targets error
    if (waiting === undefined) {
      handle(...setStreamOptions$1(targets), call);
      waiting = 0;
    }

    // Handle error
    if (error) {
      call(error);
      return
    }

    // Handle targets backpressure
    for (const target of targets) {
      const plugged = !target.write(chunk, encoding);

      if (plugged) {
        waiting++;
        target.once('drain', () => {
          waiting--;
          if (waiting <= 0) {
            call();
          }
        });
      }
    }
    if (waiting <= 0) {
      call();
    }
  }

  function final (callback) {
    next = callback;
    for (const target of targets) {
      target.end();
    }
  }

  function destroy (err, callback) {
    for (const target of targets) {
      err = destroyStream(target, err);
    }
    callback(err);
  }

  return writable({
    ...options,
    write,
    writev: undefined,
    final,
    destroy
  })
}

function pump (...args) {
  if (!isFunction(last(args))) {
    return new Promise((resolve, reject) =>
      pump(...args, err => (err ? reject(err) : resolve()))
    )
  }
  readableStream.pipeline(...args);
}

function readArray (arr, options) {
  if (!Array.isArray(arr)) {
    return readable()
  }
  let index = 0;
  return readable({
    ...options,
    objectMode: true,
    read () {
      let flowing = true;
      while (flowing && index < arr.length) {
        flowing = this.push(arr[index++]);
      }
      if (index >= arr.length) {
        this.push(null);
      }
    }
  })
}

function readify (streams, options) {
  if (streams.length <= 0) {
    return readable(options)
  } else if (streams.length === 1) {
    return streams[0]
  }

  let collector;

  let next;

  return readable({
    ...options,
    read () {
      if (collector) {
        if (next) {
          const callback = next;
          next = undefined;
          callback();
        }
        return
      }

      collector = writable({
        objectMode: true,
        write: (chunk, encoding, callback) => {
          if (this.push(chunk)) {
            callback();
          } else {
            next = callback;
          }
        }
      });

      pump(...streams, collector, err => {
        if (err) {
          this.emit('error', err);
        } else {
          this.push(null);
        }
      });
    },
    destroy (err, callback) {
      callback(destroyStream(collector, err));
    }
  })
}

function subscribe (...args) {
  if (!isFunction(last(args))) {
    return new Promise((resolve, reject) =>
      subscribe(...args, (err, result) => (err ? reject(err) : resolve(result)))
    )
  }

  const callback = args.pop();

  if (args.length <= 0) {
    callback(new Error('Expected at least one readable stream'));
    return
  }

  let value;

  const listener = data => (value = data);

  const target = last(args);

  const done = err => {
    target.removeListener('data', listener);
    callback(err, value);
  };

  if (args.length > 1) {
    pump(...args, done);
  } else {
    eos(target, { writable: false }, done);
  }

  target.addListener('data', listener);
}

function writify (streams, options) {
  if (streams.length <= 0) {
    return writable(options)
  } else if (streams.length === 1) {
    return streams[0]
  }

  // First writable stream in pipeline
  let target;

  // Current write/final callback
  let next;

  // Current pipeline error
  let error;

  // Update error and fire next if necessary
  function call (err) {
    error = err || error;
    if (next) {
      const callback = next;
      next = undefined;
      callback(error);
    }
  }

  return writable({
    ...options,
    write (chunk, encoding, callback) {
      // Initialize the pipeline
      if (!target) {
        target = streams[0];
        pump(...streams, call);
      }

      // Handle possible pipeline error
      if (error) {
        callback(error);
        return
      }

      // Perform write and handle backpressure
      if (!target.write(chunk, encoding)) {
        next = callback;
        target.once('drain', call);
      } else {
        callback();
      }
    },
    final (callback) {
      next = callback;
      target.end();
    },
    destroy (err, callback) {
      callback(destroyStream(target, err));
    }
  })
}

exports.collect = collect;
exports.duplex = duplex;
exports.duplexify = duplexify;
exports.eos = eos;
exports.finished = finished;
exports.handle = handle;
exports.isDuplex = isDuplex;
exports.isReadable = isReadable;
exports.isReadableStrictly = isReadableStrictly;
exports.isStream = isStream;
exports.isWritable = isWritable;
exports.isWritableStrictly = isWritableStrictly;
exports.mergeReadables = mergeReadables;
exports.mergeWritables = mergeWritables;
exports.pump = pump;
exports.readArray = readArray;
exports.readable = readable;
exports.readify = readify;
exports.subscribe = subscribe;
exports.transform = transform;
exports.writable = writable;
exports.writify = writify;
